




/*
 *
 *          Copyright (C) 1995, M. A. Sridhar
 *  
 *
 *     This software is Copyright M. A. Sridhar, 1995. You are free
 *     to copy, modify or distribute this software  as you see fit,
 *     and to use  it  for  any  purpose, provided   this copyright
 *     notice and the following   disclaimer are included  with all
 *     copies.
 *
 *                        DISCLAIMER
 *
 *     The author makes no warranties, either expressed or implied,
 *     with respect  to  this  software, its  quality, performance,
 *     merchantability, or fitness for any particular purpose. This
 *     software is distributed  AS IS.  The  user of this  software
 *     assumes all risks  as to its quality  and performance. In no
 *     event shall the author be liable for any direct, indirect or
 *     consequential damages, even if the  author has been  advised
 *     as to the possibility of such damages.
 *
 */



#if defined(__GNUC__)
#pragma implementation
#endif


#include "ui/event.h"
#include "ui/rectangl.h"
#include "ui/visualob.h"
#include "ui/treeview.h"

UI_Event::UI_Event(UI_EventType t, UI_VisualObject* tg,
                   UI_VisualObject* cl, NativeEventStruct* native)
{
    _type = t;
    _origin = tg;
    if (cl == NULL)
        _dest = _origin;
    else
        _dest = cl;
    if (native) {
        _nativeEvent = new NativeEventStruct;
        *_nativeEvent = *native;
    }
    else
        _nativeEvent = NULL;
    _shiftKey = FALSE;
    _ctrlKey  = FALSE;
    _metaKey  = FALSE;
}

UI_Event::UI_Event(const UI_Event& e)
{
    *this = e;
}

UI_Event::~UI_Event()
{
    if (_nativeEvent)
        delete _nativeEvent;
}



UI_Event::UI_Event ()
{
    _type = Event_Other;
    _origin = NULL;
    _dest = NULL;
    _nativeEvent = NULL;
    _shiftKey = FALSE;
    _ctrlKey  = FALSE;
    _metaKey  = FALSE;
}


void UI_Event::operator= (const UI_Event& e)
{
    _origin = e._origin;
    _dest = e._dest;
    _type = e._type;
    curPos = e.curPos;
    _key = e._key;
    param = e.param;

    _nativeEvent = new NativeEventStruct;
    if (_nativeEvent && e._nativeEvent)
        *(NativeEventStruct*) _nativeEvent = *(NativeEventStruct*)
            e._nativeEvent;
}
      


// void UI_Event::SetPosition (const UI_Point& pos)
// {
//     curPos.Origin (pos);
// }

bool UI_Event::Dispatch ()
{
    return FALSE; // Default implementation
}


static char* _EventType[] = {
    "Event_None",
    "Event_FirstSoftEvent",     
    "Event_ChildCreated",       
    "Event_CloseDown",          
    "Event_MakeInterface",      
    "Event_Quit",               
    "Event_Select", 
    "Event_Other",
    "Event_MouseMove",
    "Event_LButtonPress",
    "Event_MButtonPress",       
    "Event_RButtonPress",     
    "Event_LButtonRelease",   
    "Event_MButtonRelease",   
    "Event_RButtonRelease",   
    "Event_LButtonDblClk",
    "Event_MButtonDblClk",       
    "Event_RButtonDblClk",
    "Event_Scroll",            
    "Event_FinishScroll",      
    "Event_ScrollToBegin",     
    "Event_ScrollToEnd",       
    "Event_ScrollForwardLine", 
    "Event_ScrollBackwardLine",
    "Event_ScrollForwardPage", 
    "Event_ScrollBackwardPage",
    "Event_ScrollToPosition",  
    "Event_KeyTyped",          
    "Event_KeyDown",           
    "Event_KeyUp",             
    "Event_GetFocus",          
    "Event_LoseFocus",         
    "Event_Paint",             
    "Event_GraphicsExpose",    
    "Event_Reconfigure",       
    "Event_ViewEnter",          
    "Event_ViewLeave",          
    "Event_Iconify",            
    "Event_Deiconify",          
    "Event_StrEdChanged",       
    "Event_TreeViewNodeCollapsed", 
    "Event_TreeViewNodeExpanded"
};


static char* _KeyTypeAsChars [] = {
    "Key_F1",
    "Key_F2",
    "Key_F3",
    "Key_F4",
    "Key_F5",
    "Key_F6",
    "Key_F7",
    "Key_F8",
    "Key_F9",
    "Key_F10",
    "Key_F11",
    "Key_F12",

    "Key_Backspace",
    "Key_Insert",
    "Key_Delete",
    "Key_Home",
    "Key_End",
    "Key_PageUp",
    "Key_PageDown",
    "Key_LeftArrow",
    "Key_UpArrow",
    "Key_RightArrow",
    "Key_DownArrow",
    0
};


CL_String UI_Event::AsString () const
{
    CL_String s;
    short lastEvent = sizeof (_EventType) / sizeof (const char*);
    s.AssignWithFormat
        ("%s from id %d (%s) to id %d (%s)",
         _type >= Event_None && _type <= lastEvent ? _EventType[_type]
         : "Unknown type",
         _origin ? _origin->ViewID() : -1000,
         _origin ? _origin->ClassName() : "??",
         _dest ? _dest->ViewID() : -1000,
         _dest ? _dest->ClassName() : "??");
    if (_type == Event_Reconfigure)
        s += " shape " + curPos.AsString();
    else if (_type == Event_Paint)
        s += " invalid rectangle " + curPos.AsString();
    else if (_type == Event_MouseMove)
        s += " position " + curPos.Origin().AsString();
    else if (_type == Event_Scroll)
        s += " amount " + CL_String (param);
    else if (_type == Event_KeyDown || _type == Event_KeyTyped) {
        short key = Key ();
        s += " key ";
        if (key < UI_KeyboardEvent::Key_F1 ||
            key > UI_KeyboardEvent::Key_DownArrow)
            s += "'" + CL_String ((char) key, 1) + "' (0x"
                + CL_Integer(key).InRadix (16) + ")";
        else
            s += _KeyTypeAsChars [key - UI_KeyboardEvent::Key_F1];
    }
#if defined(__MS_WINDOWS__) || defined(__MS_WIN32__)
    if (_nativeEvent)
        s += " (Windows message code " + CL_Integer
            (_nativeEvent->message).InRadix (16) + ")";
#elif defined(__OS2__)
    if (_nativeEvent)
        s += " (OS/2 message code " + CL_Integer
            (_nativeEvent->msg).InRadix (16) + ")";
#endif
    return s;
}





// ============================ MouseEvent ============================


UI_MouseEvent::UI_MouseEvent ()
{
}
    
UI_MouseEvent::UI_MouseEvent (UI_EventType type,UI_VisualObject* origin,
                              const UI_Point& position,
                              UI_VisualObject* dest,
                              NativeEventStruct* nativeEvent)
    : UI_Event (type, origin, dest, nativeEvent)
{
    curPos = UI_Rectangle (position, 0, 0);
}


bool UI_MouseEvent::Dispatch ()
{
    if (!_origin)
        return FALSE;
    bool b = FALSE;
    switch (_type) {
    case Event_LButtonPress:
        b = _origin->ButtonDown (curPos.Origin(), UIM_Left,
                                 _shiftKey, _ctrlKey);
        break;
        
    case Event_LButtonRelease:
        b = _origin->ButtonUp (curPos.Origin(), UIM_Left);
        break;

    case Event_LButtonDblClk:
        b = _origin->DoubleClick (curPos.Origin(), UIM_Left);
        break;
        
    case Event_MButtonPress:
        b = _origin->ButtonDown (curPos.Origin(), UIM_Middle,
                                 _shiftKey, _ctrlKey);
        break;

    case Event_MButtonRelease:
        b = _origin->ButtonUp (curPos.Origin(), UIM_Middle);
        break;

    case Event_MButtonDblClk:
        b = _origin->DoubleClick (curPos.Origin(), UIM_Middle);
        break;

    case Event_RButtonPress:
        b = _origin->ButtonDown (curPos.Origin(), UIM_Right,
                                 _shiftKey, _ctrlKey);
        break;

    case Event_RButtonRelease:
        b = _origin->ButtonUp (curPos.Origin(), UIM_Right);
        break;

    case Event_RButtonDblClk:
        b = _origin->DoubleClick (curPos.Origin(), UIM_Right);
        break;

    case Event_MouseMove:
        b = _origin->MouseMove (curPos.Origin());
        break;

    default:
        CL_Error::Warning ("?? UI_MouseEvent::Dispatch: "
                           "Bad mouse event type %s", AsString().AsPtr());
        break;
    }
    return b;

}




// ============================ KeyboardEvent ========================


UI_KeyboardEvent::UI_KeyboardEvent ()
{
}
    
UI_KeyboardEvent::UI_KeyboardEvent (UI_EventType type,UI_VisualObject* origin,
                                    UI_VisualObject* dest,
                                    NativeEventStruct* nativeEvent)
    : UI_Event (type, origin, dest, nativeEvent)
{
}


bool UI_KeyboardEvent::Dispatch ()
{
    if (!_origin)
        return FALSE;
    bool b = FALSE;
    if (_type ==  Event_KeyTyped)
        b = _origin->KeyTyped (_key);
    return b;
}


// ============================ ScrollEvent ========================




UI_ScrollEvent::UI_ScrollEvent ()
{
}
    
UI_ScrollEvent::UI_ScrollEvent (UI_EventType type,UI_VisualObject* origin,
                                UI_VisualObject* dest,
                                NativeEventStruct* nativeEvent)
    : UI_Event (type, origin, dest, nativeEvent)
{
}


bool UI_ScrollEvent::Dispatch ()
{
    return FALSE;
}




// ============================ FocusEvent ========================



UI_FocusEvent::UI_FocusEvent ()
{
}
    
UI_FocusEvent::UI_FocusEvent (UI_EventType type,UI_VisualObject* origin,
                              UI_VisualObject* dest,
                              NativeEventStruct* nativeEvent)
    : UI_Event (type, origin, dest, nativeEvent)
{
}


bool UI_FocusEvent::Dispatch ()
{
    if (!_origin)
        return FALSE;
    bool b = FALSE;
    if (_type == Event_GetFocus)
        b = _origin->GetFocus ();
    else if (_type == Event_LoseFocus)
        b = _origin->LoseFocus ();
    return b;
}



// ============================ IconifyEvent ========================


UI_IconifyEvent::UI_IconifyEvent ()
{
}
    
UI_IconifyEvent::UI_IconifyEvent (UI_EventType type,UI_VisualObject* origin,
                                  UI_VisualObject* dest,
                                  NativeEventStruct* nativeEvent)
    : UI_Event (type, origin, dest, nativeEvent)
{
}


bool UI_IconifyEvent::Dispatch ()
{
    if (!_origin)
        return FALSE;
    bool b = FALSE;
    if (_type == Event_Iconify)
        b = _origin->Iconify ();
    else if (_type == Event_Deiconify)
        b = _origin->Deiconify ();
    return b;
}




// ============================ EnterLeaveEvent ========================



UI_EnterLeaveEvent::UI_EnterLeaveEvent ()
{
}
    
UI_EnterLeaveEvent::UI_EnterLeaveEvent (UI_EventType type,
                                        UI_VisualObject* origin,
                                        UI_VisualObject* dest,
                                        NativeEventStruct* nativeEvent)
    : UI_Event (type, origin, dest, nativeEvent)
{
}


bool UI_EnterLeaveEvent::Dispatch ()
{
    if (!_origin)
        return FALSE;
    bool b = FALSE;
    if (_type == Event_ViewEnter)
        b = _origin->ViewEnter (curPos.Origin());
    else if (_type == Event_ViewLeave)
        b = _origin->ViewLeave (curPos.Origin());
    return b;
}



// ============================ ReconfigureEvent ========================



UI_ReconfigureEvent::UI_ReconfigureEvent ()
{
    _type = Event_Reconfigure;
}
    
UI_ReconfigureEvent::UI_ReconfigureEvent
    (UI_VisualObject* origin,
     UI_VisualObject* dest, NativeEventStruct* nativeEvent)
    : UI_Event (Event_Reconfigure, origin, dest, nativeEvent)
{
}


bool UI_ReconfigureEvent::Dispatch ()
{
    return _origin ? _origin->Reconfigure (curPos) : FALSE;
}






// ============================ PaintEvent ========================



#include <iostream.h>
UI_PaintEvent::UI_PaintEvent ()
{
    _type = Event_Paint;
}
    
UI_PaintEvent::UI_PaintEvent (UI_VisualObject* origin,
                              UI_VisualObject* dest,
                              NativeEventStruct* nativeEvent)
    : UI_Event (Event_Paint, origin, dest, nativeEvent)
{
}


bool UI_PaintEvent::Dispatch ()
{
    return _origin ? _origin->Paint (curPos) : FALSE;
}



// ============================ SelectEvent ========================



UI_SelectEvent::UI_SelectEvent ()
{
    _type = Event_Select;
}
    
UI_SelectEvent::UI_SelectEvent (UI_VisualObject* origin,
                                UI_VisualObject* dest,
                                NativeEventStruct* nativeEvent)
    : UI_Event (Event_Select, origin, dest, nativeEvent)
{
}


bool UI_SelectEvent::Dispatch ()
{
    return _origin ? _origin->Select () : FALSE;
}



// ============================ CloseDownEvent ========================


UI_CloseDownEvent::UI_CloseDownEvent (UI_VisualObject* origin,
                                      UI_VisualObject* dest,
                                      NativeEventStruct* nativeEvent)
    : UI_Event (Event_CloseDown, origin, dest, nativeEvent)
{
}


bool UI_CloseDownEvent::Dispatch ()
{
    if (_origin)
        _origin->CloseDown ();
    return FALSE;
}


// ============================ TreeViewEvent ========================



    
UI_TreeViewEvent::UI_TreeViewEvent (UI_EventType type,
                                    UI_VisualObject* origin,
                                    long nodeId,
                                    UI_VisualObject* dest,
                                    NativeEventStruct* nativeEvent)
    : UI_Event (type, origin, dest, nativeEvent)
{
    _nodeId = nodeId;
}


bool UI_TreeViewEvent::Dispatch ()
{
    UI_TreeView* view = CL_CAST_PTR (UI_TreeView*, _origin);
    if (!view)
        return FALSE;
    bool b = FALSE;
    if (_type == Event_TreeViewNodeCollapsed)
        b = view->NodeCollapsed (_nodeId);
    else if (_type == Event_TreeViewNodeExpanded)
        b = view->NodeExpanded (_nodeId);
    return b;
}



// ============================ SoftEvent ========================



UI_SoftEvent::UI_SoftEvent ()
{
}
    
UI_SoftEvent::UI_SoftEvent (UI_EventType type,UI_VisualObject* origin,
                            UI_VisualObject* dest,
                            NativeEventStruct* nativeEvent)
    : UI_Event (type, origin, dest, nativeEvent)
{
}


bool UI_SoftEvent::Dispatch ()
{
    return FALSE;
}

