



/*
 *
 *          Copyright (C) 1995, M. A. Sridhar
 *  
 *
 *     This software is Copyright M. A. Sridhar, 1995. You are free
 *     to copy, modify or distribute this software  as you see fit,
 *     and to use  it  for  any  purpose, provided   this copyright
 *     notice and the following   disclaimer are included  with all
 *     copies.
 *
 *                        DISCLAIMER
 *
 *     The author makes no warranties, either expressed or implied,
 *     with respect  to  this  software, its  quality, performance,
 *     merchantability, or fitness for any particular purpose. This
 *     software is distributed  AS IS.  The  user of this  software
 *     assumes all risks  as to its quality  and performance. In no
 *     event shall the author be liable for any direct, indirect or
 *     consequential damages, even if the  author has been  advised
 *     as to the possibility of such damages.
 *
 */

#if defined(__GNUC__)
#pragma implementation
#endif


// Support for event handling under OS/2
#include "ui/cntroler.h"
#include "ui/composit.h"
#include "ui/timer.h"
#include "ui/menu.h"
#include "ui/stred.h"
#include "ui/treeview.h"

#include <string.h>

static struct {
    ulong         message;
    UI_EventType  yaclEvent;
} TransTable [] = {
    WM_BUTTON1DOWN,         Event_LButtonPress,
    WM_BUTTON2DOWN,         Event_MButtonPress,
    WM_BUTTON3DOWN,         Event_RButtonPress,
    WM_BUTTON1UP,           Event_LButtonRelease,
    WM_BUTTON2UP,           Event_MButtonRelease,
    WM_BUTTON3UP,           Event_RButtonRelease,
    WM_BUTTON1DBLCLK,       Event_LButtonDblClk,
    WM_BUTTON2DBLCLK,       Event_MButtonDblClk,
    WM_BUTTON3DBLCLK,       Event_RButtonDblClk,
    WM_CHAR,                Event_KeyTyped,
    WM_CLOSE,               Event_CloseDown,
    WM_DESTROY,             Event_Quit,
    WM_MENUEND,             Event_LoseFocus,
    WM_MENUSELECT,          Event_GetFocus,
    WM_MOVE,                Event_Reconfigure,
    WM_MOUSEMOVE,           Event_MouseMove,
    WM_PAINT,               Event_Paint,
    WM_SETFOCUS,            Event_GetFocus,
    WM_SIZE,                Event_Reconfigure,
    0,                      Event_Other
};



bool UI_Controller::_DoOneEvent (NativeEventStruct& qMsg)
{
    UI_Event* e = TranslateNativeEvent (qMsg);
    if (e->_origin &&
        e->_origin->WindowClass() != NULL &&
        // The window class is null for menu items. The menu item events
        // are dispatched from the WindowProc.
        e->_origin->WindowClass() != _YACLWindowClassName &&
        // ---------------------^^^^ ------------------
        // Note that we compare pointers, not strings. We cannot
        // compare strings, because OS/2's built-in window classes
        // return invalid pointers as window classes. This check is
        // needed because events targeted at a YACLWindow are dispatched
        // from YACLWindowProc.
        e->_origin->WindowClass() != _YACLSimpleClassName &&
        (e->_type != Event_KeyTyped || e->_key != '\011') &&
        // Another hack, to prevent tabs from being dispatched. They
        // should be dispatched from the WindowProcs.
        (!_eventFilter || _eventFilter->Execute (*e)))
        DispatchNativeEvent (*e);
    bool ret = (_termination && _termination->Execute (*e)) ? FALSE : TRUE;
    DispatchSoftEvents ();
    delete e;
    return ret;
}


bool UI_Controller::ProcessNativeEvents ()
{
    NativeEventStruct qMsg;
    while (_root && WinGetMsg (_hab, &qMsg, NULLHANDLE, 0, 0)) {
        WinDispatchMsg (_hab, &qMsg);
        if (!_DoOneEvent (qMsg))
            return FALSE;
//         if (_deIconifyPending == qMsg.hwnd) {
//             // We need to delay dispatching the de-iconify event until
//             // after the window is visible. Otherwise, if the event
//             // handler puts up a window, the application subsequently
//             // hangs under OS/2. THIS DOESN'T WORK!
//             UI_VisualObject* view = (*this)[qMsg.hwnd];
//             if (view) {
//                 UI_Event evt (Event_Deiconify, view, view);
//                 DispatchNativeEvent (evt);
//             }
//             _deIconifyPending = 0;
//         }
    }
    return TRUE;
}

void UI_Controller::FlushEventQueue ()
{
    NativeEventStruct qMsg;
    while (_root && WinPeekMsg (_hab, &qMsg, NULLHANDLE, 0, 0, PM_REMOVE));
}



void UI_Controller::DispatchPendingEvents ()
{
    DispatchSoftEvents ();
    NativeEventStruct qMsg;
    while (_root && WinPeekMsg (_hab, &qMsg, NULLHANDLE, 0, 0, PM_REMOVE)) {
        WinDispatchMsg (_hab, &qMsg);
        if (!_DoOneEvent (qMsg))
            break;
    }
}



bool UI_Controller::DispatchNativeEvent ( UI_Event& e )
{
    if (!_root)
        return TRUE;
    switch (e.Type()) {
    case Event_MouseMove:
        if ( _current != e.Destination () ) {
            // The mouse entered a different visual obj
            if ( _current ) {
                UI_Event leave (Event_ViewLeave, _current);
                DispatchEvent (&leave);
            }
            _current = e.Destination ();
            UI_Event enter (Event_ViewEnter, e.Destination ());
            DispatchEvent (&enter);
        }
        break;

    default:
        break;
    }
    return DispatchEvent (&e);
}


static short KeyTranslate (short keyCode)
{
    static struct {
        int virtKey;
        int yaclKey;
    } codes [] = {
        {VK_F1,    UI_KeyboardEvent::Key_F1},
        {VK_F2,    UI_KeyboardEvent::Key_F2},
        {VK_F3,    UI_KeyboardEvent::Key_F3},
        {VK_F4,    UI_KeyboardEvent::Key_F4},
        {VK_F5,    UI_KeyboardEvent::Key_F5},
        {VK_F6,    UI_KeyboardEvent::Key_F6},
        {VK_F7,    UI_KeyboardEvent::Key_F7},
        {VK_F8,    UI_KeyboardEvent::Key_F8},
        {VK_F9,    UI_KeyboardEvent::Key_F9},
        {VK_F10,   UI_KeyboardEvent::Key_F10},
        {VK_F11,   UI_KeyboardEvent::Key_F11},
        {VK_F12,   UI_KeyboardEvent::Key_F12},

        {VK_BACKSPACE,  UI_KeyboardEvent::Key_Backspace},
        {VK_INSERT, UI_KeyboardEvent::Key_Insert},
        {VK_DELETE, UI_KeyboardEvent::Key_Delete},
        {VK_HOME,   UI_KeyboardEvent::Key_Home},
        {VK_END,    UI_KeyboardEvent::Key_End},
        {VK_PAGEUP, UI_KeyboardEvent::Key_PageUp},
        {VK_PAGEDOWN, UI_KeyboardEvent::Key_PageDown},
        {VK_LEFT,   UI_KeyboardEvent::Key_LeftArrow},
        {VK_UP,     UI_KeyboardEvent::Key_UpArrow},
        {VK_RIGHT,  UI_KeyboardEvent::Key_RightArrow},
        {VK_DOWN,   UI_KeyboardEvent::Key_DownArrow},
        {0, 0}
    };

    int i = 0;
    for (; codes[i].virtKey != 0; i++)
        if (keyCode == codes[i].virtKey)
            return (short) codes[i].yaclKey;
    return keyCode;
}


UI_Event* UI_Controller::TranslateNativeEvent (NativeEventStruct& msg)
{
    UI_Event* retEvent = NULL;
    short i;
    for (i = 0; TransTable[i].message != 0; i++) {
        if ( TransTable [i].message == msg.msg ) {
            break;
        }
    }
    UI_EventType eventType  = TransTable [i].yaclEvent;
    UI_VisualObject* view = (UI_VisualObject*) _visualObjMap[msg.hwnd];
    switch (msg.msg) {
    case WM_PAINT: {
        retEvent = new UI_PaintEvent (view, view, &msg);
        RECTL rect;
        if (view) {
            if (WinQueryUpdateRect (view->_handle, &rect)) {
                long x = rect.xLeft;
                long y = view->_shape.Height() - rect.yTop - 1;
                long w = rect.xRight - rect.xLeft + 1;
                long h = rect.yTop   - rect.yBottom + 1;
                retEvent->curPos = UI_Rectangle (x, y, w, h);
            }
            else
                retEvent->curPos = UI_Rectangle (0, 0, view->_shape.Width(),
                                         view->_shape.Height());
        }
        break;
    }
        
    case WM_MENUEND: {
        UI_VisualObject* v = (UI_VisualObject*)
            _visualObjMap [(ulong) msg.mp2];
        if (!v)
            break;
        UI_Menu* menu = NULL;
        if (v->WindowClass() == WC_MENU)
            menu = (UI_Menu*) v;
        else {
            UI_MenuItem* itm = (UI_MenuItem*) v;
            menu = &(itm->Container());
        }
        if (menu) {
            UI_MenuItem* focusItem = menu->Focus();
            if (focusItem) {
                UI_FocusEvent evt (Event_LoseFocus, focusItem, focusItem);
                DispatchEvent (&evt);
            }
            menu->MoveFocusTo (NULL);
        }
        retEvent = new UI_Event (Event_Other, NULL, NULL, &msg);
        break;
    }
        
    case WM_MENUSELECT: {
        UI_VObjCollection* parent = (UI_VObjCollection*)
            _visualObjMap[msg.hwnd];
        UI_ViewID id = SHORT1FROMMP (msg.mp1);
        UI_MenuItem* itm = parent && id >= 1 && id < 32700 ?
            (UI_MenuItem*)(*parent)[id] : (UI_MenuItem*) NULL;
        // OS/2 is very picky about view id's. The above check for id range
        // between 1 and 32700 is so that rogue MENUSELECT messages do not
        // cause problems.
        if (itm) {
            UI_Menu& menu = itm->Container();
            UI_MenuItem* focusItem = menu.Focus();
            if (focusItem) {
                UI_FocusEvent evt (Event_LoseFocus, focusItem, focusItem);
                DispatchEvent (&evt);
            }
            menu.MoveFocusTo (itm);
        }
        retEvent = new UI_FocusEvent (Event_GetFocus, itm, itm, &msg);
        break;
    }
        
    case WM_CHAR: {
        retEvent = new UI_KeyboardEvent (eventType, view, view, &msg);
        retEvent->_key = KeyTranslate (SHORT1FROMMP(msg.mp2));
        // PM gives me two WM_CHAR messages (See p. 452 of Petzold's book),
        // so I ignore one of them:
        if (((ulong) msg.mp1) & KC_KEYUP || retEvent->_key == 0)
            retEvent->_type = Event_Other;
        break;
    }
        
    case WM_BUTTON1DBLCLK:
    case WM_BUTTON2DBLCLK:
    case WM_BUTTON3DBLCLK:
    case WM_BUTTON1DOWN:
    case WM_BUTTON2DOWN:
    case WM_BUTTON3DOWN:
    case WM_BUTTON1UP:
    case WM_BUTTON2UP:
    case WM_BUTTON3UP: {
        POINTL point = msg.ptl;
        WinMapWindowPoints (HWND_DESKTOP, msg.hwnd, &point, 1);
        RECTL rect;
        WinQueryWindowRect (msg.hwnd, &rect);
        retEvent = new UI_MouseEvent
            (eventType, view,
             UI_Point (point.x, rect.yTop - rect.yBottom + 1 - point.y),
             view, &msg);
        break;
    }

    case WM_SETFOCUS:
        retEvent = new UI_FocusEvent (eventType, view, view, &msg);
        retEvent->_type = SHORT1FROMMP (msg.mp2)
            ? Event_GetFocus: Event_LoseFocus;
        break;

    case WM_TIMER: {
        retEvent = new UI_Event (eventType, NULL, NULL, &msg);
        ulong id = SHORT1FROMMP (msg.mp1);
        UI_Timer::DoAlarm (id);
        break;
    }
        
    case WM_COMMAND: {
        UI_ViewID id = SHORT1FROMMP(msg.mp1);
        UI_VObjCollection* parent = (UI_VObjCollection*)
            _visualObjMap[msg.hwnd];
        view = parent ? (*parent)[id] : (UI_VisualObject*) NULL;
        retEvent = new UI_SelectEvent (view, view, &msg);
        break;
    }

    case WM_CONTROL: {
        UI_ViewID id = SHORT1FROMMP(msg.mp1);
        HWND h = WinWindowFromID (msg.hwnd, id);
        view = (UI_VObjCollection*) _visualObjMap[h];
        ushort code = SHORT2FROMMP(msg.mp1);
        //  if (code == CBN_EFCHANGE || CBN_LBSELECT)
        //      retEvent->_type = Event_Select;
        //  else
        //      retEvent->_type = (code == BN_CLICKED) ? Event_Select
        //          : Event_LButtonDblClk;
        // The CBN_ messages seem to come before the WM_COMMAND (or is it
        // WM_CONTROL), in fact, even before the selection actually changes.
        // So we can't ask the listbox what its selection is, because it
        // hasn't changed yet.
        if (code == EN_CHANGE) {
            UI_WindowClass cls = view->WindowClass();
            if (cls == WC_ENTRYFIELD || cls == WC_MLE) {
                // This is a message from a StringEditor
                UI_StringEditor* edit = (UI_StringEditor*) view;
                edit->Update();
            }
            retEvent = new UI_Event (eventType, view, view, &msg);
            break;
        }
        switch (code) {
        case CN_EMPHASIS: {
            long nodeId = -1;
            HWND h = WinWindowFromID (msg.hwnd, id);
            view = (*this)[h];
            if (view) {
                UI_TreeView* tView = (UI_TreeView*) view;
                nodeId = tView->NodeIdFromMessage (msg);
                // OS/2 seems to send two of these messages, one to the
                // old node and one to the new. We want to force that only
                // the new one gets the select event.
                long selection = tView->Selection();
                if (nodeId != selection)
                    retEvent = new UI_TreeViewEvent (Event_Select,
                                                     tView, nodeId,
                                                     tView, &msg);
            }
            if (!retEvent)
                retEvent = new UI_Event (eventType, view, view, &msg);
                // Make sure we create a non-null Event
            break;
        }

        case CN_COLLAPSETREE: {
            long nodeId = -1;
            HWND h = WinWindowFromID (msg.hwnd, id);
            view = (*this)[h];
            if (view)
                nodeId = ((UI_TreeView*) view)->NodeIdFromMessage (msg);
            retEvent = new UI_TreeViewEvent (Event_TreeViewNodeCollapsed,
                                             view, nodeId, view, &msg);
            break;
        }

        case CN_EXPANDTREE: {
            long nodeId = -1;
            HWND h = WinWindowFromID (msg.hwnd, id);
            view = (*this)[h];
            if (view)
                nodeId = ((UI_TreeView*) view)->NodeIdFromMessage (msg);
            retEvent = new UI_TreeViewEvent (Event_TreeViewNodeExpanded,
                                             view, nodeId, view, &msg);
            break;
        }

        case BN_CLICKED:
            retEvent = new UI_SelectEvent (view, view, &msg);
            break;

        default:
            retEvent = new UI_Event (eventType, NULL, NULL, &msg);
            break;

        }
        break;
    }
        
    case WM_MOUSEMOVE: {
        short x = SHORT1FROMMP (msg.mp1);
        short y = SHORT2FROMMP (msg.mp1);
        if (view)
            y = view->_shape.Height() - y; // Adjust for the fact that
                                           // OS/2's origin reckoning is
                                           // bottom left, not top left
        retEvent = new UI_MouseEvent (Event_MouseMove, view, UI_Point (x, y),
                                      view, &msg);
        break;
    }

    case WM_MOVE: {
        retEvent = new UI_ReconfigureEvent (view, view, &msg);
        if (!view)
            break;
        SWP swp;
        UI_ViewHandle vHandle = msg.hwnd;
        WinQueryWindowPos (vHandle, &swp);
        UI_ViewID vID = WinQueryWindowUShort (vHandle, QWS_ID);
        if (vID == FID_CLIENT) // It's part of a framed window
            vHandle = WinQueryWindow (vHandle, QW_PARENT); // Get the frame
        POINTL newOrigin;
        newOrigin.x = newOrigin.y = 0;
        HWND parentHandle  = WinQueryWindow (vHandle, QW_PARENT);
        WinMapWindowPoints (vHandle, parentHandle, &newOrigin, 1);
        retEvent->curPos =  UI_Rectangle
            (newOrigin.x,
             _YACLWindowHeight (parentHandle) - newOrigin.y - swp.cy,
             swp.cx, swp.cy);
        // DEBUG CODE:
//         CL_Error::Warning ("WM_MOVE view '%s' to %s\n", view->ClassName(),
//                            retEvent->curPos.AsString().AsPtr());
        break;
    }
        
    case WM_SIZE: {
        retEvent = new UI_ReconfigureEvent (view, view, &msg);
        RECTL rect;
        WinQueryWindowRect (msg.hwnd, &rect);
        if (view) {
            UI_Rectangle shape = view->Shape();
            retEvent->curPos =  UI_Rectangle (shape.Left(), shape.Top(),
                                      rect.xRight, rect.yTop);
        }
        // DEBUG CODE:
//         CL_Error::Warning ("WM_SIZE view '%s' to %s\n",
//                            view ? view->ClassName() : "",
//                            retEvent->curPos.AsString().AsPtr());
        break;
    }

    case WM_MINMAXFRAME: {
        retEvent = new UI_IconifyEvent (Event_Other, view, view, &msg);
        // Use Event_Other, because a WM_SIZE is sent which takes care
        // of the reconfigure event.
        PSWP pSwp = (PSWP) msg.mp1;
        if (view && pSwp) {
            if (pSwp->fl & SWP_MINIMIZE)
                retEvent->_type = Event_Iconify;
            else if (pSwp->fl & SWP_RESTORE) {
                _deIconifyPending = msg.hwnd;
                retEvent->_type = Event_Deiconify;
            }
        }
        break;
    }
        
    case WM_VSCROLL:
    case WM_HSCROLL: {
        UI_EventType scrollEvt = Event_Other;
        HWND parent = WinWindowFromID (msg.hwnd, SHORT1FROMMP(msg.mp1));
        view = (UI_VisualObject*) _visualObjMap [parent];
        switch (SHORT2FROMMP (msg.mp2)) {
        case SB_LINEDOWN:
            scrollEvt = Event_ScrollForwardLine;
            break;

        case SB_PAGEDOWN:
            scrollEvt = Event_ScrollForwardPage;
            break;

        case SB_LINEUP:
            scrollEvt = Event_ScrollBackwardLine;
            break;

        case SB_PAGEUP:
            scrollEvt = Event_ScrollBackwardPage;
            break;

        case SB_ENDSCROLL:
            scrollEvt = Event_FinishScroll;
            break;

        case SB_SLIDERPOSITION:
            scrollEvt = Event_ScrollToPosition;
            break;

        case SB_SLIDERTRACK:
            scrollEvt = Event_Scroll;
            break;
        }
        retEvent = new UI_ScrollEvent (scrollEvt, view, view, &msg);
        retEvent->param = SHORT1FROMMP(msg.mp2);
        break;
    }

    case WM_CLOSE:
        retEvent = new UI_CloseDownEvent (view, view, &msg);
        break;
        
    default:
        retEvent = new UI_Event (eventType, view, view, &msg);
        retEvent->curPos.Origin (UI_Point (msg.ptl.x, msg.ptl.y));
        break;
    }
    *(NativeEventStruct*)(retEvent->_nativeEvent) = msg;
    retEvent->_shiftKey = WinGetKeyState (HWND_DESKTOP, VK_SHIFT) < 0
        ? TRUE : FALSE;
    retEvent->_ctrlKey  = WinGetKeyState (HWND_DESKTOP, VK_CTRL ) < 0
        ? TRUE : FALSE;
    retEvent->_metaKey  = WinGetKeyState (HWND_DESKTOP, VK_ALT  ) < 0
        ? TRUE : FALSE;

    return retEvent;
}


void UI_Controller::_MakeOS2Interface (const UI_Event& e)
{
    UI_VisualObject* origin = e.Origin ();
    if (origin->MakeVisualElement ())
        Register (origin);
    origin->_PrivateInitialize ();
    origin->Initialize ();
    if (origin->WindowClass() == _YACLWindowClassName &&
        origin->ViewHandle () > 0) {
        // Force an initial paint event to be sent
        UI_ViewHandle h = origin->ViewHandle();
        RECTL rect;
        WinQueryWindowRect (h, &rect);
        WinInvalidateRect  (h, &rect, TRUE);
    }
}

// 
// static short _PendingEvents (UI_EventType t, UI_VisualObject* w)
// {
//     short count = 0;
//     UI_Controller& ctrlr = YACLApp()->Controller();
//     for (long i = ctrlr._eventQueue.Size() - 1; i >= 0; i--) {
//         UI_Event* e = (UI_Event*) ctrlr._eventQueue(i);
//         if (e->Type() == t && e->Origin()->Parent() == v &&
//             e->Origin()->WindowClass() != NULL) {
//             // Don't count menu items with null window class
//             count++;
//             CL_Error::Warning ("cls %s id %d EventsPending %d\n",
//                                e->Origin()->ClassName(),
//                                e->Origin()->ViewID(), count);//DEBUG
//         }
//     }
//     return count;
// }


MRESULT EXPENTRY YACLWindowProc (HWND hWnd, ULONG msg, MPARAM p1, MPARAM p2)
{
    MRESULT result = 0;
    UI_Controller& ctrlr = YACLApp()->Controller();
    switch (msg) {
    case WM_ERASEBACKGROUND: {
//         HPS hps = (HPS) p1;
//         RECTL* rect = (RECTL*) p2;
//         UI_VisualObject* v = ctrlr[hWnd];
//         long color = v ? v->Background().NativeForm() : SYSCLR_WINDOW;
//         WinFillRect     (hps, rect, color);
//         result = (MRESULT) FALSE; 
        // Return TRUE to cause PM to paint the window background
        // in SYSCLR_WINDOW
        result = (MRESULT) TRUE;
        break;
    }

    case WM_CLOSE:
        break;

        
    case WM_MOUSEMOVE: {
        if (ctrlr._inWaitState)
            ctrlr.SetCurrentCursor (ctrlr._defaultCursor);
        else {
            UI_VisualObject* v = ctrlr[hWnd];
            if (v) {
                UI_Cursor& cursor = v->Cursor();
                if (cursor != UICursor_Default) {
                    ctrlr.SetCurrentCursor (v->Cursor());
                    break;
                }
            }
        }
        result = WinDefWindowProc (hWnd, msg, p1, p2);
        break;
    }

        
    case WM_PAINT:
        // Don't do anything: the processing is done by VisualObject's Paint
        // method.
        break;

    default:
        result = WinDefWindowProc (hWnd, msg, p1, p2);
        break;
    }
    QMSG qMsg;
    WinQueryPointerPos (HWND_DESKTOP, &qMsg.ptl);
    qMsg.hwnd = hWnd;
    qMsg.msg  = msg;
    qMsg.mp1  = p1;
    qMsg.mp2  = p2;
    qMsg.time = 0;
        // WM_COMMAND events arrive both in this WindowProc and in
        // ProcessNativeEvent. Those from menu items are dispatched from
        // here, and the rest are dispatched from ProcessNativeEvent.
    
    UI_Event* e = ctrlr.TranslateNativeEvent (qMsg);
    if (e->Origin()) {
        bool disp = (qMsg.msg == WM_COMMAND) ?
            (SHORT1FROMMP(p2) == CMDSRC_MENU) : FALSE;
        // disp is now TRUE if it's a menu selection event
        if (!disp) {
            disp = qMsg.msg != WM_COMMAND && qMsg.msg != WM_VSCROLL &&
                qMsg.msg != WM_HSCROLL;
        }
        
        // disp is now TRUE if it's either a menu selection event or a paint
        // event that must be dispatched
        if (disp &&
            (!ctrlr._eventFilter || ctrlr._eventFilter->Execute (*e)))
            ctrlr.DispatchNativeEvent (*e);
    }
    delete e;
    if (qMsg.msg == WM_PAINT)
        WinValidateRect (hWnd, NULL, FALSE); // Assume that the Paint method
                                             // did its job. Without this
                                             // call I'll be flooded with
                                             // WM_PAINT messages
    return result;
}



MRESULT EXPENTRY YACLBmpBtnProc (HWND hWnd, ULONG msg, MPARAM p1, MPARAM p2)
{
    // Here's a bit of OS/2 weirdness: the default processing for WM_CHAR
    // involves sending the message to the parent! the result is that the
    // parent dispatches the event, and then this BmpBtnProc does, so we
    // have duplicate notification. So I call WinDefWindowProc only if
    // it's not a WM_CHAR message:
    MRESULT result =  (msg == WM_CHAR) ? (MRESULT) TRUE
        : WinDefWindowProc (hWnd, msg, p1, p2);
    UI_Controller& ctrlr = YACLApp()->Controller();
    QMSG qMsg;
    WinQueryPointerPos (HWND_DESKTOP, &qMsg.ptl);
    qMsg.hwnd = hWnd;
    qMsg.msg  = msg;
    qMsg.mp1  = p1;
    qMsg.mp2  = p2;
    qMsg.time = 0;
    UI_Event* e = ctrlr.TranslateNativeEvent (qMsg);
    if (e->Origin()) {
        if (!ctrlr._eventFilter || ctrlr._eventFilter->Execute (*e))
            ctrlr.DispatchNativeEvent (*e);
    }
    delete e;
    return result;
}



void UI_Controller::_AbortPMApp (const char* msg)
{
    PERRINFO  pErrInfoBlk;
    PSZ       pszOffSet;
    void      stdprint(void);
    PSZ  pszErrMsg;

    DosBeep (100, 10);
    DosBeep (440, 110);
    if ((pErrInfoBlk = WinGetErrorInfo (_hab)) != NULL) {
        pszOffSet = ((PSZ) pErrInfoBlk) + pErrInfoBlk->offaoffszMsg;
        pszErrMsg = ((PSZ )pErrInfoBlk) + *((PSHORT)pszOffSet);
        CL_String sMsg = CL_String("YACL: ") + msg + CL_String (" failed:\n")
            + pszErrMsg;
        WinMessageBox
            (HWND_DESKTOP, HWND_DESKTOP,  sMsg.AsPtr(), "YACL FATAL", 0,
             MB_MOVEABLE | MB_CUACRITICAL | MB_CANCEL );
        WinFreeErrorInfo (pErrInfoBlk);
    }
}

short UI_Controller::Initialize ()
{
    if ((_hab = WinInitialize (0)) == NULLHANDLE)
        _AbortPMApp ("WinInitialize");
    if ((_hmq = WinCreateMsgQueue (_hab, 0)) == NULLHANDLE) {
        _AbortPMApp ("WinCreateMsgQueue");
        WinTerminate (_hab);
        return FALSE;
    }
    if (!WinRegisterClass (_hab, _YACLWindowClassName, YACLWindowProc,
                           CS_SYNCPAINT | CS_MOVENOTIFY | CS_SIZEREDRAW, 0)) {
        _AbortPMApp ("WinRegisterClass");
        WinDestroyMsgQueue (_hmq);
        WinTerminate (_hab);
        return FALSE;
    }
    if (!WinRegisterClass (_hab, _YACLSimpleClassName, YACLBmpBtnProc,
                           CS_SYNCPAINT | CS_MOVENOTIFY | CS_SIZEREDRAW, 0)) {
        _AbortPMApp ("WinRegisterClass Simple");
        WinDestroyMsgQueue (_hmq);
        WinTerminate (_hab);
        return FALSE;
    }
    return TRUE;
}

