

/*
 *
 *          Copyright (C) 1995, M. A. Sridhar
 *  
 *
 *     This software is Copyright M. A. Sridhar, 1995. You are free
 *     to copy, modify or distribute this software  as you see fit,
 *     and to use  it  for  any  purpose, provided   this copyright
 *     notice and the following   disclaimer are included  with all
 *     copies.
 *
 *                        DISCLAIMER
 *
 *     The author makes no warranties, either expressed or implied,
 *     with respect  to  this  software, its  quality, performance,
 *     merchantability, or fitness for any particular purpose. This
 *     software is distributed  AS IS.  The  user of this  software
 *     assumes all risks  as to its quality  and performance. In no
 *     event shall the author be liable for any direct, indirect or
 *     consequential damages, even if the  author has been  advised
 *     as to the possibility of such damages.
 *
 */




#if defined(__GNUC__)
#pragma implementation
#endif


#if defined(__GNUC__) && __GNUC_MINOR__ >= 7
#    include <string.h>  // Without this, the X includes barf
#endif
#include <X11/Xlib.h>
#include <X11/X.h>


#include "base/bytstrng.h"

#include "ui/stddlg.h"
#include "ui/applic.h"
#include "ui/cntroler.h"
#include "ui/dialog.h"
#include "ui/pushbtn.h"
#include "ui/label.h"
#include <iostream.h>
#include "ui/dsplsurf.h"
#include "ui/color.h"

#include "ui/support/x_yacl/picture.h"
#include "ui/support/x_yacl/bitmaps/info.bmp"
#include "ui/support/x_yacl/bitmaps/warning.bmp"
#include "ui/support/x_yacl/bitmaps/error.bmp"
#include "ui/support/x_yacl/bitmaps/questhead.bmp"

#include "ui/stred.h"
#include "ui/strview.h"
#include "ui/strseq.h"
#include "io/directry.h"

#include "io/pathname.h"

enum {
    ID_SELECTIONLABEL       =    238
  , ID_SELECTION            =    239

  , ID_FILTERLABEL          =    240
  , ID_FILTER               =    241

  , ID_DIRECTORYLABEL       =    242
  , ID_DIRECTORY            =    243

  , ID_FILESLABEL           =    244
  , ID_FILES                =    245

  , ID_BTNOK                =    246
  , ID_BTNFILTER            =    247
  , ID_BTNCANCEL            =    248
} ;

UI_ViewDescriptor FileDialogDesc[] = {
    {View_Label,               ID_FILTERLABEL,  
     10, 15, 75, 15, FALSE,   "Filter", NULL}

  , {View_StringEditor,        ID_FILTER,
     10, 40, 280, 24, TRUE,    "", NULL}

  , {View_Label,               ID_DIRECTORYLABEL, 
     10, 80, 130, 15, FALSE,   "Directories", NULL}

  , {View_StringViewSingle, ID_DIRECTORY, 
     10, 105, 130, 140, TRUE,  "", NULL}

  , {View_Label,               ID_FILESLABEL, 
     160, 80, 130, 15, FALSE,    "Files", NULL}

  , {View_StringViewSingle, ID_FILES, 
     160, 105, 130, 140, TRUE,  "", NULL}

  , {View_Label,               ID_SELECTIONLABEL,  
     10, 265, 90, 15, FALSE,   "Selection", NULL}

  , {View_StringEditor,        ID_SELECTION,
     10, 290, 280, 24, TRUE,    "", NULL}

  , {View_PushButton, ID_BTNOK, 
     30, 342, 60, 30, TRUE,  "Ok", NULL}

  , {View_PushButton, ID_BTNFILTER, 
     120, 342, 60, 30, TRUE,  "Filter", NULL}

  , {View_PushButton, ID_BTNCANCEL, 
     210, 342, 60, 30, TRUE,  "Cancel", NULL}

  , {View_None, 0, 0}

} ;

class UI_FSDialog: public UI_Dialog {

    // A private class.

public:
    UI_FSDialog (UI_CompositeVObject* parent,
                 const UI_Rectangle&, CL_String, CL_String) ;
    CL_String Selection() ;

protected:
    bool HandleChildEvent (const UI_Event& e) ;

    void _InitLabels() ;
    void _SetDirectoryFiles(CL_String) ;

    bool _FileSelect() ;
    bool _DirectorySelect() ;
    bool _Filter() ;

    CL_String _currFileSel ;
    CL_String _currDirSel ;
    CL_String _currSelection ;
    const char* _pathSep ;
    
} ;


const short StdDialogWidth = 300, StdDialogHeight = 110;

class StandardDialog: public UI_Dialog {

    // A private class.
    
public:
    StandardDialog (UI_CompositeVObject* parent, const UI_Rectangle& shape)
        : UI_Dialog (parent, NULL, shape) {};

protected:
    bool HandleChildEvent (const UI_Event& e) {return TRUE ;} ;
    // Don't send events up the tree.
    
} ;


UI_ViewID UI_StandardDialog (const char* message,
                             const char* box_title,
                             UI_CompositeVObject* parent,
                             UI_StandardDialogStyle style,
                             UI_StandardDialogIconStyle icon)
{
    // Center the dialog in the parent
    short width = StdDialogWidth, height = StdDialogHeight;
    short x = 0, y = 0, pw, ph;
    UI_Rectangle shape;
    UI_Application* theApp = YACLApp();
    if (parent) {
        UI_Rectangle& parentShape = parent->Shape();
        pw = parentShape.Width();
        ph = parentShape.Height();
        x = parentShape.Left();
        y = parentShape.Top();
    }
    else {
        UI_Rectangle parentShape = theApp->ScreenRect();
        pw = parentShape.Width();
        ph = parentShape.Height();
    }
    shape = UI_Rectangle (x + maxl ((pw - width)/2, 0),
                          y + maxl ((ph - height)/2, 0), width, height);
    StandardDialog* dlg = new StandardDialog (parent, shape);
    dlg->Title() = box_title;
    UI_Font& font = dlg -> Font() ;
    long txtWidth = font.TextWidth(message) ;
   
    char* icon_bits ;
    int icon_width, icon_height ;

    switch (icon) {
        case UIS_Information: {
            icon_bits   = info_bits ;
            icon_width  = info_width ;
            icon_height = info_height ;
            break ;
        }
        case UIS_Warning: {
            icon_bits   = warning_bits ;
            icon_width  = warning_width ;
            icon_height = warning_height ;
            break ;
        }
        case UIS_Error: {
            icon_bits   = error_bits ;
            icon_width  = error_width ;
            icon_height = error_height ;
            break ;
        }
        case UIS_Question: {
            icon_bits   = questhead_bits ;
            icon_width  = questhead_width ;
            icon_height = questhead_height ;
            break ;
        }
        default: {
            icon_bits   = info_bits ;
            icon_width  = info_width ;
            icon_height = info_height ;
            break ;
        }
    }

    long iconTextWidth = 2 * (30 + icon_width) + txtWidth ; 
    long totalWidth = 0 ;
    if (iconTextWidth < 260)
        totalWidth = 260 ;
    else
        totalWidth = iconTextWidth ;

    UI_Label* lbl = new UI_Label
        (dlg, UI_Rectangle (totalWidth/2 - txtWidth/2, 20, txtWidth, 30), 10);
    lbl->Title() = message;
    UI_PushButton* btn1, *btn2, *btn3;
    switch (style) {
    case UIS_Ok:
        btn1 = new UI_PushButton
            (dlg, UI_Rectangle (totalWidth/2 - 30,  60, 60, 30), UI_IDOK);
        btn1->Title() = "Ok";
        break;

    case UIS_OkCancel:
        btn1 = new UI_PushButton
            (dlg, UI_Rectangle (totalWidth/2 - 70,  60, 60, 30), UI_IDOK);
        btn1->Title() = "Ok";
        btn2 = new UI_PushButton
            (dlg, UI_Rectangle (totalWidth/2 + 10,  60, 60, 30),
             UI_IDCANCEL);
        btn2->Title() = "Cancel";
        break;

    case UIS_YesNo:
        btn1 = new UI_PushButton
            (dlg, UI_Rectangle (totalWidth/2 - 70,  60, 60, 30),
             UI_IDYES);
        btn1->Title() = "Yes";
        btn2 = new UI_PushButton
            (dlg, UI_Rectangle (totalWidth/2 + 10,  60, 60, 30),
             UI_IDNO);
        btn2->Title() = "No";
        dlg->AddDialogEvent (UI_IDYES, Event_Select);
        dlg->AddDialogEvent (UI_IDNO,  Event_Select);
        break;

    case UIS_YesNoCancel:
        btn1 = new UI_PushButton
            (dlg, UI_Rectangle (totalWidth/2 - 110, 60, 60, 30),
             UI_IDYES);
        btn1->Title() = "Yes";
        btn2 = new UI_PushButton
            (dlg, UI_Rectangle (totalWidth/2 -  30, 60, 60, 30), UI_IDNO);
        btn2->Title() = "No";
        btn3 = new UI_PushButton
            (dlg, UI_Rectangle (totalWidth/2 +  50, 60, 60, 30),
             UI_IDCANCEL);
        btn3->Title() = "Cancel";
        dlg->AddDialogEvent (UI_IDYES, Event_Select);
        dlg->AddDialogEvent (UI_IDNO,  Event_Select);
        break;

    }
    int icon_x, icon_y ;

    icon_x = (totalWidth/2 - txtWidth/2)/2 - icon_width/2 ; 

    //28 = (35 - 7), 35 is lbl.Top() + lbl.Height() / 2 ;
    icon_y = (28 - icon_height / 2) ; 

    UI_Rectangle iconShape(icon_x, icon_y, icon_width, icon_height) ;

    UI_Picture* bmp =  new UI_Picture(dlg, iconShape, icon_bits);

    UI_Rectangle dlg_shape = dlg -> Shape() ;
    UI_Rectangle dlgShape (dlg_shape.TopLeft(), totalWidth, dlg_shape.Height()) ;
    dlg -> Shape() = dlgShape ;

    UI_DialogEventDescriptor d = dlg->ExecuteModal ();
    YACLApp()->Destroy (dlg);
    delete bmp ;
    return d.id;
}

UI_FSDialog::UI_FSDialog (UI_CompositeVObject* parent,
                          const UI_Rectangle& shape, CL_String box_title, 
                          CL_String initialDir)
    : UI_Dialog (parent, FileDialogDesc, shape) {

    Title() = box_title ;
    _InitLabels() ;

    _pathSep = CL_Directory::PathSeparator() ;
    if (!_pathSep) 
        _pathSep = "/" ;

    long iS = initialDir.Size() ;
    if (initialDir(iS - 1) == *_pathSep)
        initialDir(iS - 1,iS - 1) = "" ;

    _SetDirectoryFiles(initialDir) ;

    AddDialogEvent (ID_BTNOK, Event_Select);
    AddDialogEvent (ID_BTNCANCEL, Event_Select);

}

void UI_FSDialog::_InitLabels() {
    UI_Label* lbl ;
    lbl = (UI_Label*) (*this)[ID_SELECTIONLABEL] ;
    lbl -> SetTextStyle(UIText_Left) ;
    lbl = (UI_Label*) (*this)[ID_FILESLABEL] ;
    lbl -> SetTextStyle(UIText_Left) ;
    lbl = (UI_Label*) (*this)[ID_DIRECTORYLABEL] ;
    lbl -> SetTextStyle(UIText_Left) ;
    lbl = (UI_Label*) (*this)[ID_FILTERLABEL] ;
    lbl -> SetTextStyle(UIText_Left) ;
}

void UI_FSDialog::_SetDirectoryFiles(CL_String currentDir) {

    CL_Directory currDir(currentDir + _pathSep) ;
    if (!currDir.IsValid())
        return ; 

    _currFileSel =  "" ;
    _currDirSel  = currentDir ;
    _currSelection = _currDirSel + _pathSep + _currFileSel ;

    UI_StringEditor* selEdit = (UI_StringEditor*) (*this)[ID_SELECTION] ;
    selEdit -> Model() = _currSelection ;

    UI_StringEditor* filtEdit = (UI_StringEditor*) (*this)[ID_FILTER] ;
    filtEdit -> Model() = (_currSelection + "*") ;

    UI_StringViewSingleSel* dirsel = 
         (UI_StringViewSingleSel*)(*this)[ID_DIRECTORY] ;

    UI_StringViewSingleSel* filesel = 
         (UI_StringViewSingleSel*)(*this)[ID_FILES] ;

    UI_StringSequence& dirseq = (UI_StringSequence&)dirsel -> Model() ;
    UI_StringSequence& fileseq = (UI_StringSequence&)filesel -> Model() ;

    CL_StringSequence strseq = currDir.FileNames() ;

    int size = strseq.Size() ;

    UI_StringSequence dir_seq;
    UI_StringSequence file_seq;

    CL_String currFile ;

    for (int i = 0; i < size; i++) {
        currFile = currentDir + _pathSep + strseq(i) ;
        if (CL_Directory::IsDirectory(currFile))
            dir_seq.Add(currFile) ;
        else
            file_seq.Add(strseq(i)) ;
    }
    dir_seq.Sort() ;
    file_seq.Sort() ;

    dirseq = dir_seq ;
    fileseq = file_seq ;
}

bool UI_FSDialog::_FileSelect() {
    UI_StringViewSingleSel* filesel = 
         (UI_StringViewSingleSel*)(*this)[ID_FILES] ;
    int sel = filesel -> Selection() ;
    UI_StringSequence& fileseq = (UI_StringSequence&)filesel -> Model() ;

    if (sel < 0 || sel > fileseq.Size())
        return TRUE ;

    _currFileSel = fileseq(sel) ;
    _currSelection = _currDirSel + _pathSep + _currFileSel ;
  
    UI_StringEditor* selEdit = (UI_StringEditor*) (*this)[ID_SELECTION] ;
    selEdit -> Model() = _currSelection ;
    return TRUE ;
}

bool UI_FSDialog::_DirectorySelect() {
    UI_StringViewSingleSel* dirsel = 
         (UI_StringViewSingleSel*)(*this)[ID_DIRECTORY] ;
  
    int sel = dirsel -> Selection() ;
    UI_StringSequence& dirseq = (UI_StringSequence&)dirsel -> Model() ;
    if (sel <= 0 || sel >= dirseq.Size())
        return TRUE ;

    CL_String initstr = dirseq(sel) ;
    if (sel == 1) {
        long l = initstr.Size() ;
        initstr(l - 3, l - 1) = "" ; 
        const char* p = initstr.AsPtr() ;
        l = initstr.Size() ;
         
        long i ;
        for (i = l - 1; i > 0; i--)
            if (p[i] == *_pathSep)
                break ;

        initstr(i,l) = "" ;
    }

    _SetDirectoryFiles(initstr) ;
    return TRUE ;
}

bool UI_FSDialog::_Filter() {
    UI_StringEditor* filtStr = (UI_StringEditor*) (*this)[ID_FILTER] ;
    CL_String filterString ;
    filterString = filtStr -> Model() ;

    int len = filterString.Size() ;
    long i ;
    const char* p = filterString.AsPtr(); 
    for (i = len - 1; i >= 0; i--)
        if (p[i] == *_pathSep)
            break ;

    //this is important for filtering; 
    //but for now, we are not doing actual filter, so discard it.
    filterString(i,len) = "" ; 

    _SetDirectoryFiles(filterString) ;
    return TRUE ;
}

CL_String UI_FSDialog::Selection() {
    UI_StringEditor* selEdit = (UI_StringEditor*) (*this)[ID_SELECTION] ;
    _currSelection = selEdit -> Model() ;
    return _currSelection ;
}

bool UI_FSDialog::HandleChildEvent(const UI_Event& e) {

    if (e.Type() == Event_Select) {
        long id = e.Origin() -> ViewID() ; 
        switch(id) {
            case ID_DIRECTORY:
                _DirectorySelect() ;
                break;
            case ID_FILES:
                _FileSelect() ;
                break;
            case ID_BTNFILTER:
                _Filter() ;
                break;
            default:
                break;
        }
    }
    else if ( (e.Type() == Event_KeyTyped) && (e.Key() == -243) ) {
        if (e.Origin() -> ViewID() == ID_FILTER) {
            UI_PushButton* filtButton = (UI_PushButton*) (*this)[ID_BTNFILTER] ;
            UI_Event evt(Event_Select,filtButton,filtButton) ;
            _Controller -> DispatchEvent(&evt) ;
        }
        else if (e.Origin() -> ViewID() == ID_SELECTION) {
            UI_PushButton* selectButton = (UI_PushButton*) (*this)[ID_BTNOK] ;
            UI_Event evt(Event_Select,selectButton,selectButton) ;
            _Controller -> DispatchEvent(&evt) ;
        }
    }
    return TRUE ;
}

CL_String UI_FileSelectDialog (UI_CompositeVObject* parent,
                               const char* initialDir,
                               const char* box_title,
                               const char* typeMask)
{
    short width = 300 , height = 400;
    short x = 0, y = 0, pw, ph;
    UI_Rectangle shape;

    shape = UI_Rectangle (60, 60, width, height) ;
    UI_FSDialog * fsDlg = new UI_FSDialog (parent, shape, box_title,
                                           initialDir) ;
    UI_DialogEventDescriptor d = fsDlg -> ExecuteModal () ;
    if (d.id == ID_BTNOK) {
        CL_String retStr = fsDlg->Selection() ;
        YACLApp() -> Destroy(fsDlg) ;
        return retStr ;
    }
    else if (d.id == ID_BTNCANCEL) {
        YACLApp() -> Destroy(fsDlg) ;
        return "" ;
    }
    else
        return "" ;
}



bool UI_FontDialog (UI_CompositeVObject* parent, UI_FontDesc& desc)
{
    return FALSE;
}

